# -*- coding: utf-8 -*-
"""
Run like this: first run in this directory:
    git clone https://github.com/malb/lattice-estimator.git

    sage: import sys
    sage: sys.path.append('./lattice-estimator')
    sage: attach("estimator_eurocrypt.py")
    sage: %time results = runall()

"""
# Copyright Amaury Pouly and Yixin Shen
# Licensed under the Apache License, Version 2.0, see LICENSE for details.
# SPDX-License-Identifier: Apache-2.0

from sage.all import sqrt, log, exp, tanh, coth, e, pi, RR, ZZ

from estimator.cost import Cost
from estimator.lwe_parameters import LWEParameters
from estimator.reduction import delta as deltaf
from estimator.reduction import RC, ReductionCost
from estimator.reduction import cost as cost_bkz
from estimator.conf import red_cost_model as red_cost_model_default
from estimator.util import local_minimum, early_abort_range
from estimator.io import Logging
from estimator.schemes import (
    Kyber512,
    Kyber768,
    Kyber1024
)
from estimator.nd import NoiseDistribution
import itertools
from enum import Enum
from dataclasses import dataclass
import tabulate

# Precision of computations.
myRR = RealField(200)

# Gaussian sampler to use
class Sampler(Enum):
    GPV = 1 # GPV, improved by BLPRS13
    # The complexity of the MCMC algorith essentially depends on
    # alpha^2 = |b1| / s
    # see https://eprint.iacr.org/2019/660.pdf (note that s=sqrt(2pi)sigma
    # using their notations)
    # If we take alpha = 2 then the complexity is essentially polynomial (and
    # does not need to assume the GSA)
    # If we take alpha = 1 then complexity becomes quite high without assuming the GSA.
    # If we take alpha = 1/sqrt(2) then we really need to assume the GSA to use the fact
    # that the |bi*| decreases exponentially fast to get a good complexity.
    MCMC_TWO = 2 # alpha=2
    MCMC_ONE = 3 # alpha = 1
    MCMC_SQRT_HALF = 4 # alpha = 1/sqrt(2) which is what we suggest in the paper.

class ModulusSwitching(Enum):
    Off = 1
    EstimateOn = 2 # estimate complexity with ideal modulus switching

@dataclass
class Parameters:
    sampler: Sampler
    mod_switching: ModulusSwitching

# This is the "exact" formula for the complexity of T_MCMC(L,s) where:
# * n is the dimension of the lattice L,
# * we assume that we have a BKZ-beta reduced basis,
# * the volume of the lattice L is red_vol^n.
def mcmc_sampling_complexity_exact(
    n,
    beta,
    red_vol,
    s
):
    res = myRR(1)
    # The complexity is the product (for i=1,..n) for rho_{|b_i*|/s}(Z)
    # where b_i* is the i-th GS vector. With the GSA, we have
    # - |b_i*|=delta_beta^{-2(-1)}*|b_1|
    # - |b_1|=vol^{1/n}*delta_beta^n
    delta_beta = deltaf(beta)
    print("delta = ", delta_beta)
    for i in range(1, n):
        s_i = (red_vol*delta_beta**(n-2*(i-1)) / s)
        # approximate rho_{s_i}(Z) with a few of the biggest terms
        res *= sum([myRR(exp(-pi*x**2/s_i**2)) for x in range(-10,11)])
    return res

# Same as mcmc_sampling_complexity_exact() but use an approximate formula
# only valid when |b_1|<=2s.
def mcmc_sampling_complexity(
    n,
    beta,
    red_vol,
    s
):
    # The complexity is the product (for i=1,..n) for rho_{|b_i*|/s}(Z)
    # where b_i* is the i-th GS vector. With the GSA, we have
    # - |b_i*|=delta_beta^{-2(i-1)}*|b_1|
    # - |b_1|=vol^{1/n}*delta_beta^n
    # Since computing the product is expensive, we use the fact that
    # |b_i|/s will be rather small so rho_{|b_i|/s}(Z) will be close to 1,
    # hence, the log of the complexity is very close to for i=1,..,n
    # of log(rho_{|b_i*|/s}(Z)) ~= log(1+2*exp(-pi*s^2/|b_i*|^2)) ~= 2*exp(-pi*s^2/|b_i*|^2)
    # Still, computing the sum of all such terms is expensive so we replace it
    # by the integral when sum for i=1,...,infinity. There is an exact expression:
    # int(exp(-alpha*x^n),n=0..infinity)=Ei(1, alpha)/ln(x)
    delta_beta = deltaf(beta)
    b1 = red_vol * delta_beta**n
    # Check constraint on s, allow for small error in comparison due to floating point approx.
    assert b1/(2*s) <= 1.01, "b1={} needs to be smaller than 2s={}".format(myRR(b1), myRR(2*s))
    alpha = pi * s**2 / b1**2
    res = myRR(ln(1+2*exp(-alpha)) + 2 * exp_integral_e(1, alpha)/ln(delta_beta**4))
    # The following formula get mores precise as we add terms but it is more expensive:
    # for k in range(1, 5):
    #     res += -(-2)**k / k * exp_integral_e(1, k * alpha)/ln(delta_beta**4)
    return myRR(exp(res))

# Estimate the cost of the attack given all the parameters.
def compute_cost(
    scheme,
    m,
    n_guess,
    beta,
    red_cost_model,
    params,
    quiet = True,
    use_exact_mcmc = False
):
    assert m > scheme.n, "n={}, m={}, beta={}".format(scheme.n, m, beta)
    # n = n_dual + n_guess
    n_dual = scheme.n - n_guess
    # compute the two lower bounds on s
    use_mcmc = False
    if params.sampler == Sampler.GPV:
        klein_factor = sqrt(ln(2*scheme.n+4)/pi)
    elif params.sampler == Sampler.MCMC_TWO:
        klein_factor = sqrt(2)
        use_mcmc = True
    elif params.sampler == Sampler.MCMC_ONE:
        klein_factor = 1
        use_mcmc = True
    elif params.sampler == Sampler.MCMC_SQRT_HALF:
        klein_factor = 1/2
        use_mcmc = True
    else:
        assert False, "unsupported"
    if not quiet: print("klein factor: {}".format(klein_factor))
    mins1 = deltaf(beta)**m * scheme.q**(n_dual / m - 1) * klein_factor
    mins2 = 1 / (scheme.q**(1-scheme.n/m)/sqrt(e) - 2*scheme.Xe.stddev)
    if not quiet: print("min s: {} and {}".format(myRR(mins1), myRR(mins2)))

    # always choose the smallest possible value
    s = RR(max(mins1, mins2))

    # if we use MCMC, the sampling complexity depends on the choice of s
    sampling_complexity = 1
    if use_mcmc:
        # sample from D_{L_q^\perp(A),qs}: the lattice has volume q^n_dual
        compl_fn = mcmc_sampling_complexity_exact if use_exact_mcmc else mcmc_sampling_complexity
        sampling_complexity = compl_fn(m, beta, scheme.q**(n_dual / m), scheme.q * s)

    # now delta is fixed
    delta = exp(-m * s**2 * scheme.Xe.stddev**2  / 2) / 100
    if not quiet: print("delta=", myRR(delta))
    # avoid dealing with huge number, if delta is too small the complexity will be bad anyway
    if delta < RR(1e-300):
        return None
    # now N is fixed
    N = scheme.n * log(scheme.q) / delta**2

    # compute the cost of computing a short vector of norm delta_0^m*vol(L)^(1/m) by BKZ
    bkz_cost = cost_bkz(red_cost_model, beta, m)
    
    if params.mod_switching == ModulusSwitching.Off:
        # real formula:
        guessing_comp = scheme.q**n_guess
    elif params.mod_switching == ModulusSwitching.EstimateOn:
        # fake modulus switching (estimated complexity)
        guessing_comp = 2**n_guess
    else:
        assert False, "unsupported"

    rop = sampling_complexity * N + guessing_comp + bkz_cost["rop"]
    if not quiet: print("##### N={}, guess={}, BKZ={}".format(int(log(N)/log(2)), int(log(scheme.q**n_guess)/log(2)), int(log(bkz_cost["rop"])/log(2))))
    if not quiet: print("rop:", myRR(rop))
    cost = Cost(rop = rop, problem = scheme)
    cost["delta"] = delta
    cost["N"] = N
    cost["s"] = s
    cost["beta"] = beta
    cost["n_guess"] = n_guess
    cost["n_dual"] = n_dual
    cost["m"] = m
    cost["sampling"] = sampling_complexity
    return cost

# Search for the best possible parameters for a scheme given a scheme.
def opt_cost(scheme, max_m, red_cost_model, params):
    best_cost = None
    
    # Without modulus switching, we can also consider very small values of n_guess
    if params.mod_switching == ModulusSwitching.Off:
        max_guess = 40
        guess_step = 2
    elif params.mod_switching == ModulusSwitching.EstimateOn:
        max_guess = scheme.n
        guess_step = 20
    
    for m in range(scheme.n+1, max_m, 50):
        for n_guess in range(1, max_guess, guess_step):
            for beta in range(10, m, 20):
                this_cost = compute_cost(scheme, m, n_guess, beta, red_cost_model, params)
                if this_cost is not None and (best_cost is None or this_cost < best_cost):
                    best_cost = this_cost

    # Once we have the cost, recompute with exact MCMC complexity model.
    real_cost = compute_cost(scheme, m, best_cost['n_guess'], best_cost['beta'],
                             red_cost_model, params, use_exact_mcmc = True)
    print("sampling: approx=", best_cost['sampling'], "real=", real_cost['sampling'])

    return best_cost

ALL_SCHEMES = [
    {"name": "Kyber512", "scheme": Kyber512},
    {"name": "Kyber768", "scheme": Kyber768},
    {"name": "Kyber1024", "scheme": Kyber1024},
]

# Run the optimizer on Kyber given some parameters.
def runall_with(txt, params):
    kyber512 = opt_cost(Kyber512, 1024, RC.MATZOV.__class__(nn='list_decoding-classical'), params)
    print("Kyber512 ({}):".format(txt))
    print(kyber512)

    kyber768 = opt_cost(Kyber768, 768 * 2, RC.MATZOV.__class__(nn='list_decoding-classical'), params)
    print("Kyber768 ({}):".format(txt))
    print(kyber768)

    kyber1024 = opt_cost(Kyber1024, 2048, RC.MATZOV.__class__(nn='list_decoding-classical'), params)
    print("Kyber1024 ({}):".format(txt))
    print(kyber1024)

# Run the optimizer on Kyber with the MCMC sampler, with and without modulus switching.
def runall():
    results = {}
    for scheme in ALL_SCHEMES:
        scheme_results = {}
        for use_mod_switch in [False, True]:
            mod_switch_name = "{} modulus switching".format("with" if use_mod_switch else "no")
            params = Parameters(
                Sampler.MCMC_SQRT_HALF,
                ModulusSwitching.EstimateOn if use_mod_switch else ModulusSwitching.Off,
            )
            sc = scheme["scheme"]
            max_m = sc.m * 2
            res = opt_cost(sc, max_m, RC.MATZOV.__class__(nn='list_decoding-classical'), params)
            print("{} ({}):".format(scheme["name"], mod_switch_name))
            print(res)
            scheme_results[use_mod_switch] = res
        results[scheme["name"]] = scheme_results
    return results

EUROCRYPT_PARAMS = [
    {"name": "Kyber512", "scheme": Kyber512, "mod_switch": False, "m": 1013, "n_guess": 15, "beta": 550},
    {"name": "Kyber768", "scheme": Kyber768, "mod_switch": False, "m": 1469, "n_guess": 23, "beta": 870},
    {"name": "Kyber1024", "scheme": Kyber1024, "mod_switch": False, "m": 2025, "n_guess": 31, "beta": 1230},
    {"name": "Kyber512", "scheme": Kyber512, "mod_switch": True, "m": 763, "n_guess": 141, "beta": 390},
    {"name": "Kyber768", "scheme": Kyber768, "mod_switch": True, "m": 1169, "n_guess": 201, "beta": 610},
    {"name": "Kyber1024", "scheme": Kyber1024, "mod_switch": True, "m": 1575, "n_guess": 261, "beta": 890},
]

# Reproduce the estimates from Eurocrypt paper, does not run the optimizer but only
# the cost function
def reproduce_eurocrypt(params = EUROCRYPT_PARAMS):
    for p in params:
        use_mod_switch = p['mod_switch']
        mod_switch_name = "{} modulus switching".format("with" if use_mod_switch else "no")
        params = Parameters(
            Sampler.MCMC_SQRT_HALF,
            ModulusSwitching.EstimateOn if use_mod_switch else ModulusSwitching.Off,
        )
        res = compute_cost(p["scheme"], p["m"], p["n_guess"], p["beta"], RC.MATZOV.__class__(nn='list_decoding-classical'), params, True)
        print("{} ({}):".format(p["name"], mod_switch_name))
        print(res)

class ColumnType(Enum):
    INTEGER = 0
    FLOAT = 1
    LOG2_INTEGER = 2

COLUMNS = [
    ("attack", "rop", ColumnType.LOG2_INTEGER),
    ("m", "m", ColumnType.INTEGER),
    ("$n_\\guess$", "n_guess", ColumnType.INTEGER),
    ("$n_\\dual$", "n_dual", ColumnType.INTEGER),
    ("$\\beta$", "beta", ColumnType.INTEGER),
    ("s", "s", ColumnType.FLOAT),
]

CASES = [
    ("no modulus switching", False),
    ("with modulus switching", True)
]

ROWS = [
    ("Kyber512", "Kyber512"),
    ("Kyber768", "Kyber768"),
    ("Kyber1024", "Kyber1024"),
]

# Assume a array of the form: {scheme_name_1: result_1, ...}
# where each result_i is of the form: {case_1: subresult_1, ...}
# where subresult_i is a Cost structure
def results_table_latex(results, columns = COLUMNS, cases = CASES, rows = ROWS):
    col_headers = [col[0] for col in columns] * len(cases)
    all_rows = []
    for r in rows:
        if r[1] not in results:
            print("warning: row {} not found in results".format(r[1]))
            continue
        this_row = [r[0]]
        for case in cases:
            res = results[r[1]][case[1]]
            for col in columns:
                typ = col[2]
                if typ == ColumnType.INTEGER:
                    this_row.append(int(res[col[1]]))
                elif typ == ColumnType.FLOAT:
                    this_row.append("{:.2f}".format(res[col[1]]))
                elif typ == ColumnType.LOG2_INTEGER:
                    this_row.append(int(log(res[col[1]])/log(2)))
                else:
                    assert False, "unsupported"
        all_rows.append(this_row)

    return tabulate.tabulate(
        all_rows,
        headers=["Scheme"] + col_headers,
        tablefmt="latex_raw",
        floatfmt=".3f",
    )